#include "RayCast.h"
#include "World.h"
#include "Vector.h"
#include "Block.h"

Vec3f raycast(Vec3f origin, Vec3f direction, float radius, int &facei) {
	// From "A Fast Voxel Traversal Algorithm for Ray Tracing"
	// by John Amanatides and Andrew Woo, 1987
	// <http://www.cse.yorku.ca/~amana/research/grid.pdf>
	// <http://citeseer.ist.psu.edu/viewdoc/summary?doi=10.1.1.42.3443>
	// Extensions to the described algorithm:
	//   • Imposed a distance limit.
	//   • The face passed through to reach the current cube is provided to
	//     the callback.

	// The foundation of this algorithm is a parameterized representation of
	// the provided ray,
	//                    origin + t * direction,
	// except that t is not actually stored; rather, at any given point in the
	// traversal, we keep track of the *greater* t values which we would have
	// if we took a step sufficient to cross a cube boundary along that axis
	// (i.e. change the integer part of the coordinate) in the variables
	// tMaxX, tMaxY, and tMaxZ.

	// Cube containing origin point.
	float x = (int)(origin[0]);
	float y = (int)(origin[1]);
	float z = (int)(origin[2]);
	// Break out direction vector.
	float dx = direction[0];
	float dy = direction[1];
	float dz = direction[2];
	// Direction to increment x,y,z when stepping.
	int stepX = signum(dx);
	int stepY = signum(dy);
	int stepZ = signum(dz);
	// See description above. The initial values depend on the fractional
	// part of the origin.
	float tMaxX = intbound(origin[0], dx);
	float tMaxY = intbound(origin[1], dy);
	float tMaxZ = intbound(origin[2], dz);
	// The change in t when taking a step (always positive).
	float tDeltaX = stepX/dx;
	float tDeltaY = stepY/dy;
	float tDeltaZ = stepZ/dz;
	// Buffer for reporting faces to the callback.
	Vec3f face;

	// Avoids an infinite loop.
	if (dx == 0 && dy == 0 && dz == 0)
		Assert(0);

	// Rescale from units of 1 cube-edge to units of 'direction' so we can
	// compare with 't'.
	radius /= sqrtf(dx*dx+dy*dy+dz*dz);

	while (/* ray has not gone past bounds of world */
			(stepX > 0 ? x < MW : x >= 0) &&
			(stepY > 0 ? y < MH : y >= 0) &&
			(stepZ > 0 ? z < MD : z >= 0)) {

		// Invoke the callback, unless we are not *yet* within the bounds of the
		// world.
		if (!(x < 0 || y < 0 || z < 0 || x >= MW || y >= MH || z >= MD)) {
			//if (callback(x, y, z, blocks[x*wy*wz + y*wz + z], face))
			//	break;
			if (World::getBlock(x, y, z)) {
				if (face.y > 0) facei = (int)Top   ;
				if (face.y < 0) facei = (int)Bottom;
				if (face.x > 0) facei = (int)Right ;
				if (face.x < 0) facei = (int)Left  ;
				if (face.z > 0) facei = (int)Front ;
				if (face.z < 0) facei = (int)Back  ;
				
				return Vec3f(x, y, z);
			}
		}

		// tMaxX stores the t-value at which we cross a cube boundary along the
		// X axis, and similarly for Y and Z. Therefore, choosing the least tMax
		// chooses the closest cube boundary. Only the first case of the four
		// has been commented in detail.
		if (tMaxX < tMaxY) {
			if (tMaxX < tMaxZ) {
				if (tMaxX > radius) break;
				// Update which cube we are now in.
				x += stepX;
				// Adjust tMaxX to the next X-oriented boundary crossing.
				tMaxX += tDeltaX;
				// Record the normal vector of the cube face we entered.
				face[0] = -stepX;
				face[1] = 0;
				face[2] = 0;
			} else {
				if (tMaxZ > radius) break;
				z += stepZ;
				tMaxZ += tDeltaZ;
				face[0] = 0;
				face[1] = 0;
				face[2] = -stepZ;
			}
		} else {
			if (tMaxY < tMaxZ) {
				if (tMaxY > radius) break;
				y += stepY;
				tMaxY += tDeltaY;
				face[0] = 0;
				face[1] = -stepY;
				face[2] = 0;
			} else {
				// Identical to the second case, repeated for simplicity in
				// the conditionals.
				if (tMaxZ > radius) break;
				z += stepZ;
				tMaxZ += tDeltaZ;
				face[0] = 0;
				face[1] = 0;
				face[2] = -stepZ;
			}
		}
	}
	return Vec3f(-1, -1, -1);
}

float intbound(float s, float ds) {
	// Find the smallest positive t such that s+t*ds is an integer.
	if (ds < 0) {
		return intbound(-s, -ds);
	} else {
		s = mod(s, 1);
		// problem is now s+t*ds = 1
		return (1-s)/ds;
	}
}

int signum(int x) {
	return x > 0 ? 1 : x < 0 ? -1 : 0;
}

int signum(float x) {
	return x > 0 ? 1 : x < 0 ? -1 : 0;
}

float mod(float value, int modulus) {
	//return (value % modulus + modulus) % modulus;
	float val = (value - (int)value);
	val += 1;
	val = (val - (int)val);
	return val;
}
